#include <stdarg.h>

#include "console.h"
#include "r8c-ports.h"
#include "globals.h"
#include "fix.h"
#include "i2c.h"
#include "msec.h"
#include "leds.h"
#include "twiddle.h"

/* Most of the modules are interrupt driven.  However, the 1wire
   peripherals must be polled.  Thus, we use main() to poll them.  */

volatile byte trigger;
volatile byte p1_bits;

static unsigned char onewirebuf[20];
unsigned char *onewireval;
extern unsigned char ow_wire_therm_addr[8];

unsigned char *ow_target = ow_wire_therm_addr;

#define BUTTON_ATTENTION_SPAN	10*1000

unsigned char climate_data[4];

static void
mwait()
{
  static int last_ms = 0;
  while ((msec - last_ms) < 2);
  last_ms = msec;
}

static void
mspause(int ms)
{
  unsigned int start = msec;
  while (msec - start < ms)
    asm("");
}

static int sending_addr = 0;
#if 0
static int need_eol=0;

static void ow_eol()
{
  if (!need_eol)
    return;
  need_eol = 0;
  cput_c ('\n');
}

static int rst()
{
  int i = ow_reset();
  ow_eol();
  printf("rst%d", i);
  need_eol = 1;
  return i;
}

static int rd1()
{
  int i = ow_read ();
  printf(" r\033[33m%d\033[0m", i);
  need_eol = 1;
  return i;
}

static int rd8()
{
  int i = ow_read_byte ();
  printf(" r\033[32m%b\033[0m", i);
  need_eol = 1;
  return i;
}

static int wr8(int b)
{
  int i;
  i = ow_write_byte (b);
  if (i != b)
    printf(" w\[033[31m%b=%b\033[0m", b, i);
  else if (sending_addr)
    printf(" w\033[35m%b\033[0m", b);
  else
    printf(" w\033[34m%b\033[0m", b);
  need_eol = 1;
  return i;
}

#define ow_reset rst
#define ow_read rd1
#define ow_read_byte rd8
#define ow_write_byte wr8
#else
#define ow_eol()
#endif

void
wait_for_complete()
{
  int i = 3;
  /* Three in case we get a presence pulse.  */
  while (i)
    {
      i--;
      if (ow_read() == 0)
	{
	  /*cput_c('W');*/
	  i = 3;
	}
    }
}

int
s16(unsigned char lsb, unsigned char msb)
{
  int rv;
  rv = lsb + (int)msb * 256;
  return rv;
}

void
poll_buttons ()
{
  int i, found_buttons;
  static int prev_results = -1;

  ow_search_addr[0] = 0x01;
  ow_search_init (1);

  found_buttons = 0;
  while (ow_search_next ())
    {
      /*printf("I got %x error %d\n", ow_search_addr, 8, ow_search_error);*/

      for (i=0; i<4; i++)
	if (memcmp (ow_wire_button_addrs[i], ow_search_addr, 8) == 0)
	  {
	    /*printf("it's button %d\n", i);*/
	    found_buttons |= 1<<i;
	  }
    }
  if (ow_search_error)
    return;

  if (found_buttons)
    msec_timeout = BUTTON_ATTENTION_SPAN;
  if (prev_results != found_buttons)
    {
      prev_results = found_buttons;
      return;
    }
  status_byte = (status_byte & 0xf0) | (found_buttons & 0x0f);
}

/*#define RETURN0 {printf("\033[31m ret:%d\033[0m\n", __LINE__);return 0;}*/
#define RETURN0 {led_off(GREEN); led_on(RED); return 0;}
/*#define RETURN0 return 0;*/

int
onewire(int len, ...)
{
  int i, rv = 1;
  int do_wait = 0;
  int first=-1;
  va_list a;

  onewireval = onewirebuf;

  if (len < 0)
    {
      len = -len;
      do_wait = 1;
    }

  if (ow_reset())
    RETURN0;

#if 1
  i = ow_write_byte(0x55);
  if (i != 0x55)
    RETURN0;
  sending_addr = 1;
  for (i=0; i<8; i++)
    {
      if (ow_write_byte(ow_target[i]) != ow_target[i])
	{
	  sending_addr = 0;
	  RETURN0;
	}
    }
  sending_addr = 0;
#else
  i = ow_write_byte(0xCC);
#endif

  va_start (a, len);
  for (i=0; i<len; i++)
    {
      int c = va_arg (a, int);
      if (first == -1)
	first = c;
      if (c == -1)
	{
	  onewireval = onewirebuf + i;
	  break;
	}
      onewirebuf[i] = ow_write_byte(c);
      if (onewirebuf[i] != c)
	rv = 0;
    }
  for (; i<len; i++)
    {
      onewirebuf[i] = ow_read_byte();
    }
  if (do_wait)
    {
      /*cput_hex_byte (first);*/
      /*cput_c(' ');*/
      wait_for_complete();
      /*cput_c ('\n');*/
    }
  return rv;
}

#define REDO if (redo == 0) { redo = 1; goto write_page_again; }
#define REDO0 { REDO; RETURN0 }

int
poll_climate ()
{
  int i = 0, p, retry, redo;
  static int page[2][8];

  for (p=0; p<2; p++)
    {
      redo = 0;
    write_page_again:
      /* write page 0, Vad or Vdd */
      if (!onewire(3, 0x4e, 0x00, p&1 ? 0x08 : 0x00))
	RETURN0;

      /* read page 0 */
      if (!onewire (11, 0xbe, 0, -1))
	{
	  /*printf("can't read page 0\n");*/
	  continue;
	}
      if ((onewireval[0] & 0x08) != ((p&1) ? 0x08 : 0x00))
	{
	  /*printf("\033[33mre-writing page 0\033[0m\n");*/
	  REDO0;
	}

      /* save page 0 */
      if (!onewire(-2, 0x48, 0x00))
	REDO0;

      if (p == 0)
	/* initiate T convert */
	if (!onewire(-1, 0x44))
	  REDO0;

      /* initiate V convert */
      if (!onewire(-1, 0xb4))
	REDO0;

      /* recall page 0 */
      if (!onewire(-2, 0xb8, 0))
	REDO0;

#define MAX_RETRIES 4
      for (retry=0; retry < MAX_RETRIES; retry++)
	{
	  /* read page 0 */
	  if (!onewire (11, 0xbe, 0, -1))
	    {
	      /*printf("can't read page 0\n");*/
	      continue;
	    }

	  i = crc(onewireval, 9);
	  if (i != 0)
	    {
	      /*printf("\033[34mbad crc %x %b\033[0m\n", onewireval, 8, i);*/
	      continue;
	    }

	  if ((onewireval[0] & 0x08) != ((p&1) ? 0x08 : 0x00))
	    {
	      /*printf("oops: %x\n", onewireval, 8);*/
	      REDO;
	      RETURN0;
	    }

	  break;
	}
      if (retry == MAX_RETRIES)
	RETURN0;

      for (i=0; i<8; i++)
	page[p][i] = onewireval[i];
    }
  ow_eol();

  int traw = s16(page[0][1], page[0][2]);
  int vad = s16(page[0][3], page[0][4]);
  int vdd = s16(page[1][3], page[1][4]);

  //printf("t %w   vad %w %d   vdd %w %d\n", traw, vad, vad, vdd, vdd);

  if (vdd > 550 || vdd < 400)
    RETURN0;
  if (vad > vdd - 5)
    RETURN0;

  if (page[0][2] == 0xff && page[0][1] == 0xff)
    RETURN0;

#if 1
  led_on(GREEN);
  led_off(RED);
#endif

  FixType temp = fix_make8 (page[0][2], page[0][1]);

  temp = fix_mul (temp, F(1,0.8)); /* 9/5 */
  temp = fix_add (temp, F(32,0));

  /*double humidity1 = ((((double)vad/(double)vdd) - 0.16) / 0.0062) / (1.093 - 0.0012 * temp);*/
  FixType humidity;
  humidity = fix_div (fix_make (vad, 0), fix_make (vdd, 0));
  humidity = fix_sub (humidity, F(0,0.16));
  humidity = fix_mul (humidity, F(161,0.29032)); /* 1/0.0062 */

  FixType correction = fix_sub (F(1,0.093), fix_mul (F(0,0.0012), temp));
  humidity = fix_div (humidity, correction);

  asm("fclr I");
  climate_data[0] = page[0][2]; /* Th */
  climate_data[1] = page[0][1]; /* Tl */
  climate_data[2] = humidity.i;
  climate_data[3] = humidity.f >> 8;
  status_byte |= 0x80;
  asm("fset I");

#if 0
  printf("\033[32mTemp %w %F\tVad %w %d Vdd %w %d %F%%\033[0m\n",
	 traw, temp,
	 vad, vad, vdd, vdd, humidity);
#endif
#if 0
  {
    static char twiddlebites[] = "\\|/-";
    static int twiddle = 0;
    printf(" %c \r", twiddlebites[twiddle++]);
    twiddle &= 3;
  }
#endif

  return 1;
}

climate_tx_start()
{
  i2c_tx_buffer = climate_data;
  i2c_tx_count = 4;
  status_byte &= ~0x80;
}

static const i2c_setup_struct climate_command = {
  0, 0, climate_tx_start, 0, 0
};

#if 0
extern volatile int tx_put asm("tx_put");
extern volatile int tx_get asm("tx_get");
extern volatile int count asm("count");
extern volatile unsigned char i2c_state;
static int oput, oget, ocount, oiccr2, oi2c_state;
#endif

main()
{
  int i = 0, j, k;
  unsigned int last_poll_time = 0;
  int poll_delay = 100;

  pd1 = 0x9f; /* LEDS output, P1.7 and P1.4 are output. */
  pd3 = 0xff; /* output */
  pd4 = 0x00; /* input */

  drr = 0x0f; /* drive capacity high */

  p1_bits = 0xff;
  p1 = p1_bits;
  p3 = 0x00;

  i2c_init ();
  console_init ();
  msec_init ();
  serial_init ();
  ow_init ();
  leds_init ();
  thermocouple_init ();

  led_off(RED|GREEN|YELLOW);

  i2c_register_command (3, &climate_command);

  /* This pause ensures that the host synchronizes on our serial
     console, because resetting may have caused glitches.  */
  mspause(2);
  printf("\n\n\033[34mR8C Zone Monitor\033[0m\n");

#if 0
  printf("Sending just zeros...\n");
  while (1)
    ow_write_byte (0x00);
#endif

#if 0
  printf("Sending just ones...\n");
  while (1)
    ow_write_byte (0xff);
#endif

#if 0
  printf("Reading...\n");
  while (1)
    ow_read_byte();
#endif

#if 0
  printf("sending serial...\n");
  while (1)
    {
      if (tx_count < 5)
	{
	  cput_c('x');
	  serial_test();
	}
    }
#endif

  while (1)
    {
      int i, j;

      poll_buttons ();

      serial_send ();

#if 0
#if 1
      {
	if (tx_put != oput
	    || tx_get != oget
	    || count != ocount
	    || oi2c_state != i2c_state)
#endif
	  {
	    oput = tx_put;
	    oget = tx_get;
	    ocount = count;
	    oi2c_state = i2c_state;
	    printf("serial %w %w %w %w %b \n",
		   msec_timeout, oput, oget, ocount, oi2c_state);
#if 0
	    if (oi2c_state == 0x08)
	      led_off (RED);
	    else
	      led_on (RED);
#endif
	  }
      }
#endif

      if (msec_timeout)
	continue;

      if (msec - last_poll_time > poll_delay)
	{
	  if (poll_climate ())
	    poll_delay = 400;
	  else
	    poll_delay = 100;
	  last_poll_time = msec;

	  poll_thermocouple ();
	}
#if 1
      if (msec - last_poll_time > 50
	  && !(status_byte & 0x0f))
	led_off(RED|GREEN);
#endif

      ow_eol();

#if 0
      serial_ping();
#endif

#if 0
      if (tx_count)
	{
	  led_on(YELLOW);
	  run_tx_queue ();
	}
      else
	led_off(YELLOW);
#endif
    }
}
